/**
 * A jQuery plugin to drop files from the local file system and upload them directly into the
 * specified location.
 * 
 * @see <a href="http://www.w3.org/TR/progress-events/">W3C: Progress Events (Working Draft)</a>
 * @see <a href="http://www.w3.org/TR/XMLHttpRequest2/">W3C: XMLHttpRequest Level 2 (Working Draft)</a>
 * 
 * @author Christophe Lauret
 * @version 29 February 2012
 */

/* global jQuery */

// Wrap in a closure
(function($) {

  // Plug ins options
  var o = {};

  // Flag to indicate when the plugin is already busy loading files
  var busy = false;

  // Flag to indicate when objects are being dragged on the box.
  var dragon = false;

  // An array of files IDs being currently loaded
  var loading = [];

  // Main Plugin declaration
  $.fn.psDropFile = function(options) {

    // Extend default options
    o = $.extend({}, $.fn.psDropFile.defaults, options);

    // Drop area
    var $droparea = $(this);

    // Label
    var $label = $('<span></span>');
    $droparea.append($label);

    // File input
    var $input = $('<input name="file-0" type="file">');
    $input.bind("change", fileChange);
    $input.css({'opacity' : '0', 'position': 'absolute', 'cursor': 'pointer'});
    if ($.fn.psDropFile.multipleSupported && o.multiple) {
      $input.attr("multiple", "multiple");
    }

    // Form and iframe
    var $form = $('<form action="'+o.url+'" method="post" target="upload-results" enctype="multipart/form-data"></form>');
    var $iframe = $('<iframe name="upload-results" src="about:blank" style="display: none"></iframe>');
    $form.append($input);
    $droparea.append($form).append($iframe);
    $iframe.load(function() {
      if (busy) { $.fn.psDropFile.formUploadFinished($(this).contents()); }
    });

    // maintain the position of the file browse button under the mouse when not dragging files
    $droparea.mousemove(function(e) {
      if (dragon) { return false; }
      var p = $droparea.position();
      if (e.pageX > p.left && e.pageY > p.top && e.pageX < (p.left + $droparea.outerWidth()) && e.pageY < (p.top + $droparea.outerHeight())) {
        $input.css({left: e.pageX - $input.width() + 5, top: e.pageY - $input.height() + 5});
      } else {
        $input.css({left: 0, top: 0});
      }
    });
    $input.css({left: 0, top: 0});

    // If File Drag & Drop is supported
    if ($.fn.psDropFile.supported) {
      $label.text('Click to browse files or Drop files here...');

      // We use the DOM API instead of jQuery.bind() to get access to the dataTransfer object
      var droparea = $droparea.get(0);
      droparea.addEventListener("drop",      drop,      true);
      droparea.addEventListener("dragenter", dragenter, true);
      droparea.addEventListener("dragover",  dragover,  true);
      droparea.addEventListener("dragleave", dragleave, true);

    } else {
      $label.text('Click to browse files...');
    }

    // As required by jQuery plugin conventions
    return this;
  };

  // Default values
  $.fn.psDropFile.defaults = {
    url : "",
    method : "PUT",
    hoverClass : 'hover',
    numConcurrentUploads : 20,
    // update upload speed every second
    uploadRateRefreshTime : 1000,
    multiple: true, /* TODO multiple */
    fileNameFilter: null /* TODO regular expression to filter files to accept */
  };

  // Support Methods ==============================================================================

  // Check and record whether the browser supports the XMLHttpRequest.upload
  $.fn.psDropFile.supported = (function checkSupport() {
    var xhr = new XMLHttpRequest();
    var support = !!xhr.upload && !!window.File && !!window.FileReader && !!window.FileList;
    return support;
  })();

  // Check and record whether the browser supports the multiple attribute on file elements
  $.fn.psDropFile.multipleSupported = (function() {
    var i = document.createElement('input');
    return 'multiple' in i;
  })();

  // Utility Methods ==============================================================================

  /**
   * Generate a random ID for the given name.
   * @param name A name
   */
  function generateID(name) {
    var t = ((new Date().getTime())+'').slice(-4);
    var n = name? name.replace(/[\W\.]/g, 'x').toLowerCase() : 'anonymous';
    return  n.slice(0,2)+'-'+t+'-'+n.slice(-7,-5)+'-'+Math.floor(Math.random() * 1000);
  }

  // Generate an ID
  $.fn.psDropFile.generateID = generateID;

  // Public Callback Methods ======================================================================

  /**
   * Invoked when new files are dropped.
   * 
   * @param files The files dropped
   */
  $.fn.psDropFile.newFilesDropped = function(files) {
  };

  /**
   * Invoked when the upload for given file has been started
   * @param fileID A unique identifier for the file
   * @param file   The actual File object dropped
   */
  $.fn.psDropFile.uploadStarted = function(fileID, file) {
  };

  /**
   * Invoked when the upload for given file has been finished
   * 
   * @param fileID A unique identifier for the file
   * @param file   The actual File object dropped
   * @param time   The total time it took for the upload to finish
   * @param xhr    The corresponding XML Http Request Object 
   */
  $.fn.psDropFile.uploadFinished = function(fileID, file, time, xhr) {
  };

  /**
   * Invoked when a form upload has started.
   * 
   * @param data The data returned by the upload
   */
  $.fn.psDropFile.formUploadStarted = function() {
  };

  /**
   * Invoked when the form upload has been finished.
   * 
   * @param data The data returned by the upload
   */
  $.fn.psDropFile.formUploadFinished = function(data) {
  };

  /**
   * Invoked when the progress for given file has changed
   * 
   * @param fileID   the file index (String)
   * @param file     the file object (File)
   * @param progress the progress event for that file (ProgressEvent)
   */
  $.fn.psDropFile.fileUploadProgressUpdated = function(fileID, file, progress) {
  };

  /**
   * Invoked when the upload speed of given file has changed.
   * 
   * @param fileID      the file index (String)
   * @param file        the file object (File)
   * @param KBperSecond the number of KB per sec 
   */ 
  $.fn.psDropFile.fileUploadSpeedUpdated = function(fileID, file, KBperSecond) {
  };

  // Private Methods ==============================================================================

  /**
   * Remove the specified ID from the list of loading files.
   * @param id The file ID in the loading
   */
  function unload(id) {
    for (var i = 0; i < loading.length; i++) {
      if (loading[i] === id) { 
        loading.splice(i, 1);
        return true;
      }
    }
    return false;
  }

  /**
   * For when the file input changes (Files have been browsed and selected)
   */
  function fileChange(event) {
    event.preventDefault();
    // If we can upload them using AJAX
    if ($.fn.psDropFile.supported) {
      var files = this.files;
      uploadFiles(files);
    // Otherwise, use a good ol' form
    } else {
      busy = true;
      $.fn.psDropFile.formUploadStarted();
      this.form.submit();
    }
  }


  /**
   * Indicates the dragged object can be dropped
   * 
   * <p>Note: all implementation do not provide all attributes...
   * 
   * @param dataTransfer the DataTransfer object.
   */
  function canDrop(dataTransfer) {
    var i, f, t, files, types, size;
    if (dataTransfer) {
      // Check whether we have reached our limit
//      if ((types.length + loading.length) > o.numConcurrentUploads) { return false;}

      // If files are available check they have a length
      if (!!dataTransfer.files) {
        files = dataTransfer.files;
        // Check that the type include 'Files'
        for (i = 0; i < files.length; i++) {
          f = files[i];
          size = f.size || f.fileSize;
          if (size === 0) { return false; }
        }
      }

      // If the types are specified, ensure it contains 'Files'
      if (!!dataTransfer.types) {
        types = dataTransfer.types;
        // Check that the type include 'Files'
        for (i = 0; i < types.length; i++) {
          t = types[i];
          if (t === 'Files') { return true; }
        }
        return false;
      }
    }
    return false;
  }

  /**
   * Called when entering the Drop zone while dragging objects.
   * @param event (including 'dataTransfer')
   */
  function dragenter(event) {
    dragon = true;
    $(this).addClass(o.hoverClass);
    if (canDrop(event.dataTransfer)) {
      // Required to allow dragging
      event.stopPropagation();
      event.preventDefault();
      return false;
    } else {
      $(this).addClass('no-drop');
      return true;
    }
  }

  function dragover(event) {
    // Required to allow drag-and-drop
    event.stopPropagation();
    event.preventDefault();
    return false;
  }

  function dragleave(event) {
    dragon = false;
    $(this).removeClass(o.hoverClass).removeClass('no-drop');
    // Required to allow drag-and-drop
    event.stopPropagation();
    event.preventDefault();
    return false;
  }

  function drop(event) {
    dragon = false;
    $(this).removeClass(o.hoverClass).removeClass('no-drop');
    var dt = event.dataTransfer;
    var files = dt.files;
    // Required to allow drag-and-drop
    event.preventDefault();
    uploadFiles(files);
    return false;
  }

  /**
   * Handled change of status of XMLHttpRequest.
   * @this XMLHttpRequest
   * @param event DOM Event 
   */
  function xhrstateupdate(event) {
    if (this.readyState === 4) {
      var u = this.upload;
      var time = new Date().getTime() - u.downloadStartTime;
      var ok = unload(u.fileID);
      if (ok) {
        $.fn.psDropFile.uploadFinished(u.fileID, u.fileObj, time, this);
      }
    }
  }

  /**
   * Handler for the upload 'progress' event.
   * @this XMLHttpRequest.upload
   * @param event DOM Event 
   */
  function progress(event) {
    if (event.lengthComputable) {
      var percentage = Math.round((event.loaded * 100) / event.total);
      if (this.currentProgress != percentage) {

        this.currentProgress = percentage;
        $.fn.psDropFile.fileUploadProgressUpdated(this.fileID, this.fileObj, event);

        var elapsed = new Date().getTime();
        var diffTime = elapsed - this.currentStart;
        if (diffTime >= o.uploadRateRefreshTime) {
          var diffData = event.loaded - this.startData;
          var speed = diffData / diffTime; // in KB/sec

          $.fn.psDropFile.fileUploadSpeedUpdated(this.fileID, this.fileObj, speed);

          this.startData = event.loaded;
          this.currentStart = elapsed;
        }
      }
    }
  }

  /**
   * Main function responsible for uploading the files to the server.
   * 
   * @param files An array of files to be 
   */
  function uploadFiles(files) {
    if ($.fn.psDropFile.newFilesDropped(files) === false) { return; }
    for (var i = 0; i < files.length; i++) {
      var file = files[i];

      // ignore resources starting with a '.'
      if (/^\.(.*)$/.test(file.name || file.fileName)) {continue;}

      // Create a new XMLHttpRequest object for each file
      var xhr = new XMLHttpRequest();
      var upload = xhr.upload;
      upload.fileID = generateID(file.name || file.fileName);
      upload.fileObj = file;
      upload.downloadStartTime = new Date().getTime();
      upload.currentStart = upload.downloadStartTime;
      upload.currentProgress = 0;
      upload.startData = 0;

      // Add listeners
      upload.addEventListener("progress", progress, false);

      // Send the request
      xhr.open(o.method, o.url);
      xhr.setRequestHeader("Cache-Control", "no-cache");
      xhr.setRequestHeader("X-Requested-With", "XMLHttpRequest");
      xhr.setRequestHeader("X-File-Name", file.name || file.fileName);
      xhr.setRequestHeader("X-File-Size", file.size || file.fileSize);
      xhr.setRequestHeader("Content-Type", "binary/octet-stream");

      // Capture state 
      xhr.onreadystatechange = xhrstateupdate;

      // And send the file
      xhr.send(file);

      // Upload started
      $.fn.psDropFile.uploadStarted(upload.fileID, file);
      loading.push(upload.fileID);
    }
  }

})(jQuery);